"""
2008-05-21
This module defines a class to build an HTML tree view from a list of lists.
The HTML view format is based on:
http://www.codeproject.com/KB/scripting/dhtml_treeview.aspx
By: D.D. de Kerf

I created it to display URLs in a tree, but it will be useful for other tree
stuff as well.

This class works with MyTreeView.js on the client side

Erez Bibi
erezbibi@frontiernet.net
"""

class MyTreeView:
	"""
	The input for this Tree is a list of lists. Every item in every list is a
	node in the tree. The nodes from every list creates a brunch in the tree.
	There are no double nodes in the same place.
	If the input will be:
		['aaa', 'bbb', 'ccc']
		['aaa', 'bbb', 'ddd']
	The tree will look like:
		- aaa
			- bbb
				- ccc
				- ddd
	In addition the user can control a few elements of the tree presentation.
	"""
	def __init__ (self, data = []):
		"""
		Init all the elements to thier default value, and assign an optional
		list to view.
		"""
		self._default_fold = 0
		self._node_name = None
		self._node_prefix = None
		self._node_postfix = None
		self._group_prefix = None
		self._group_postfix = None
		self._current_path = None
		self._open_path = None
		self._tree = None
		self._tree_view = None
		self._lst_data = data
		# The user can reassign this function
		self.translate_path_func = lambda lst: '/'.join (lst)


	def SetDefaultFolding (self, level):
		"""
		Sets the initial fold level.
			0 - everything is folded, if level is more then the tree branching
			level, it will open the entire tree.
		"""
		self._default_fold = level
		self._tree_view = None

	def SetNodeDecoration (self, display, prefix = '', postfix = ''):
		"""
		HTML code for the node display (name), prefix and postfix.
		Prefix will be just before the node name, postfix will be after the
		nodes children.
		Every 'NAME' in each of the strings will be replaced by the node name as
		it is in the input lists. Every 'PATH' will be replaced by the current
		path in the tree (list of nodes above this point). It will operate the
		function self.translate_path_func on the list. This function should get
		a list and return a string. It can be modified by the user.
		"""
		self._node_name = display
		self._node_prefix = prefix
		self._node_postfix = postfix
		self._tree_view = None

	def SetGroupDecoration (self, prefix = '', postfix = ''):
		"""
		Group decoration is the HTML just befor a group of nodes begins, and
		just after a group of nodes ends.
		"""
		self._group_prefix = prefix
		self._group_postfix = postfix
		self._tree_view = None

	def SetOpenPath (self, path):
		"""
		Set a path (list) not to foled even if it is longer then the fold level
		"""
		self._open_path = path
		self._tree_view = None

	def AddItem (self, lst):
		"""
		Add a list to the data lists.
		Every item in this list will be a node, and with each item, the tree
		view will go on level deeper.
		for example: if you add the list ['aaa', 'bbb', 'ccc'] the view will be:
			- aaa
				- bbb
					* ccc
		But, a node will not apper twice in the tree, in other words, if a tree
		"brunch" is already in the view, it will not add it again, it will just
		add the rest of the nodes under it.
		"""
		self._lst_data.append (lst)

	def RemoveItem (self, lst):
		"""
		Removes a list from the tree view if exists.
		"""
		if lst in self._lst_data:
			self._lst_data.remove (lst)

	def Create (self):
		"""
		Creates the tree structure from the input lists.
		You must call this method before calling GetHtmlString, and after each
		time the input lists are changed.
		"""
		# First create a recursive list of tupels as (name, sub_list)
		self._tree = []
		self._tree_view = None
		for lst in self._lst_data:
			self._AddListToTree (self._tree, lst)

	def GetHtmlString (self):
		"""
		Returns the HTML view of the tree. Before calling Create it will
		return an empty string. It will build the HTML string only once, if
		Create wasn't called again.
		"""
		if not self._tree: return ''
		if not self._tree_view:
			# Create a HTML string from the tree
			self._tree_view = '<div>%s\n' % self._group_prefix
			self._level = 1
			self._current_path = []
			for name, val in self._tree:
				self._tree_view += self._GetHtmlNode (name, val)
			self._tree_view += '\n%s</div>' % self._group_postfix
		return self._tree_view

	### Private Functions ###
	def _AddListToTree (self, lst_target, lst_source):
		"""
		Recursive function to add a list to the tree.
		initially lst_target is the entire tree and lst_source is the data to
		add. The function go over the source list and creates nodes in the
		target list.
		"""
		if not lst_source: return	# Base Case
		# Get a sub list from target by the first name in source
		new_lst = self._GetSubList (lst_target, lst_source[0])
		if new_lst is None:
			new_lst = []	# Sub list doesn't exists so make new one
			# and add to target as a tuple (name, sub_lst)
			lst_target.append ((lst_source[0], new_lst))
		# Now work on the sub list with the rest of the source list
		self._AddListToTree (new_lst, lst_source[1:])

	def _GetSubList (self, lst, name):
		"""
		Return a sub list from the tree. all the items in the tree are tuples
		(name, sub_list). The function goes over it and looks for 'name'
		If there is no item in the list with the name 'name', it will returns
		None.
		"""
		for itm, val in lst:
			if itm == name:
				return val
		return None

	def _GetHtmlNode (self, name, val):
		"""
		This function creates a HTML node.
			name is the nodes name
			val is the nodes childrens.
		This is a recursive function with _GetHtmlGroup
		Some HTML code in the returned value is hard codded and some is user
		controlled. The hard codded HTML is for the HTML result to work with
		the javascript that goes with this tree view (MyTreeView.js).
		"""
		self._current_path.append (name)	# Add the node name to the path
		ret = '%s<a onclick="return Toggle(this)"><img src="%s" border="0"></a>%s%s%s\n' % (
			self._InsertNamePath (self._node_prefix, name),	# Node prefix
			self._GetImage (val),							# Node +/- image
			self._InsertNamePath (self._node_name, name),	# Node display
			self._GetHtmlGroup (val),						# Recursive call for nodes childes
			self._InsertNamePath (self._node_postfix, name)	# Node postfix
		)
		self._current_path.pop ()			# Remove the node name from the path
		return ret

	def _GetHtmlGroup (self, group):
		"""
		This function creates the HTML code for a group of nodes.
		This is a recursive function with _GetHtmlNode.
		"""
		# set the div layer and group prefix
		ret = '\n<div style="display: %s">%s\n' % (
			self._GetHidden (), self._group_prefix
		)
		self._level += 1			# We are going one level down the tree
		for name, val in group:		# Iterate over childes
			ret += '%s%s' % (
				'\t' * self._level,					# Indentation
				self._GetHtmlNode (name, val),		# Recursive call for nodes
			)
		self._level -= 1			# Go back up in the tree
		# Set group postfix and close div layer
		ret += '%s</div>\n' % self._group_postfix
		return ret

	def _InsertNamePath (self, string, name):
		"""
		Replace 'NAME' in a string with the nodes name.
		Replace 'PATH' with the node path
		"""
		if not string: return ''
		count = string.count ('NAME')
		if count > 0:
			string = string.replace ('NAME', '%s')
			lst = tuple ([name]) * count
			string = string % lst
		count = string.count ('PATH')
		if count > 0:
			string = string.replace ('PATH', '%s')
			lst = tuple ([self.translate_path_func (self._current_path)]) * count
			string = string % lst
		return string

	def _GetImage (self, lst):
		"""
		Get an image next to a node (+,-,.)
		'lst' is the nodes sub nodes. If it is empty the node is a leaf.
		Otherwise it depend on our folding level.
		"""
		if not lst:	img = 'leaf.gif'
		elif self._ShouldFold (): img = 'plus.gif'
		else: img = 'minus.gif'
		return img

	def _GetHidden (self):
		"""
		Return a string to set to the "display" style of a div tag.
		If the node should be foldded, it hides it.
		"""
		display = ''
		if self._ShouldFold (): display = 'none'
		return display

	def _ShouldFold (self):
		"""
		Return True if the current node should be folded
		Node shouldn't be folded if it is "before" the folding level, or if it
		is on the path defined in self._open_path
		"""
		if self._level < self._default_fold: return False
		if not self._open_path or not self._current_path: return True
		if len (self._current_path) > len (self._open_path): return True
		return tuple (self._open_path[:len (self._current_path)]) \
			!= tuple (self._current_path)


if __name__ == '__main__':
	# Lists to base the tree on
	lst = []
	lst.append (('aaa', 'bbb', 'ccc', 'ddd'))
	lst.append (('aaa', 'bbb', 'fff'))
	lst.append (('aaa', 'bbb', 'eee'))
	lst.append (('foo', 'bar'))

	# Make the tree
	tree = MyTreeView (lst)
	tree.SetNodeDecoration ('<a href="MyPage?url=PATH">NAME</a>', '<LI>', '</LI>')
	tree.SetGroupDecoration ('<UL style="list-style-type: none">', '</UL>')
	tree.SetDefaultFolding (0)
	# Path to leave open (un folded)
	tree.SetOpenPath (('aaa', 'bbb', 'fff'))

	tree.RemoveItem (('aaa', 'bbb', 'eee'))
	tree.AddItem (('aaa', 'ccc', 'ddd'))
	tree.Create ()

	## Uncomment this to create a working HTML code (save result buffer as .html)
	#print '<script language="javascript" src="MyTreeView.js" />\n'
	#print tree.GetHtmlString ()
	#exit (0)

	assert tree.GetHtmlString ().strip () == \
"""<div><UL style="list-style-type: none">
<LI><a onclick="return Toggle(this)"><img src="minus.gif" border="0"></a><a href="MyPage?url=aaa">aaa</a>
<div style="display: "><UL style="list-style-type: none">
		<LI><a onclick="return Toggle(this)"><img src="minus.gif" border="0"></a><a href="MyPage?url=aaa/bbb">bbb</a>
<div style="display: "><UL style="list-style-type: none">
			<LI><a onclick="return Toggle(this)"><img src="plus.gif" border="0"></a><a href="MyPage?url=aaa/bbb/ccc">ccc</a>
<div style="display: none"><UL style="list-style-type: none">
				<LI><a onclick="return Toggle(this)"><img src="leaf.gif" border="0"></a><a href="MyPage?url=aaa/bbb/ccc/ddd">ddd</a>
<div style="display: none"><UL style="list-style-type: none">
</UL></div>
</LI>
</UL></div>
</LI>
			<LI><a onclick="return Toggle(this)"><img src="leaf.gif" border="0"></a><a href="MyPage?url=aaa/bbb/fff">fff</a>
<div style="display: "><UL style="list-style-type: none">
</UL></div>
</LI>
</UL></div>
</LI>
		<LI><a onclick="return Toggle(this)"><img src="plus.gif" border="0"></a><a href="MyPage?url=aaa/ccc">ccc</a>
<div style="display: none"><UL style="list-style-type: none">
			<LI><a onclick="return Toggle(this)"><img src="leaf.gif" border="0"></a><a href="MyPage?url=aaa/ccc/ddd">ddd</a>
<div style="display: none"><UL style="list-style-type: none">
</UL></div>
</LI>
</UL></div>
</LI>
</UL></div>
</LI>
<LI><a onclick="return Toggle(this)"><img src="plus.gif" border="0"></a><a href="MyPage?url=foo">foo</a>
<div style="display: none"><UL style="list-style-type: none">
		<LI><a onclick="return Toggle(this)"><img src="leaf.gif" border="0"></a><a href="MyPage?url=foo/bar">bar</a>
<div style="display: none"><UL style="list-style-type: none">
</UL></div>
</LI>
</UL></div>
</LI>

</UL></div>"""

	print "Test Passed"

